/**
 * pcl.js - v1.13.0-alpha.1
 *
 * Copyright (c) 2022 pcl.js Authors
 * SPDX-License-Identifier: MIT
 */

import { Vector3, Scene, PerspectiveCamera, WebGLRenderer, BufferGeometry, PointsMaterial, Float32BufferAttribute, Points, AxesHelper, GridHelper, Color } from 'three';
import { OrbitControls } from 'three/examples/jsm/controls/OrbitControls';
import { PCDLoader } from 'three/examples/jsm/loaders/PCDLoader';

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function getCenter(geometry) {
    var _a;
    geometry.computeBoundingBox();
    const center = new Vector3();
    (_a = geometry.boundingBox) === null || _a === void 0 ? void 0 : _a.getCenter(center);
    return center;
}

class PointCloudViewer {
    constructor(canvas, width = 200, height = 200) {
        this.scene = new Scene();
        this.camera = new PerspectiveCamera();
        this.clouds = [];
        this.cloudProperties = {
            sizeAttenuation: false,
            size: 1,
            color: '#FFF',
            id: 'point-cloud',
        };
        // private function
        this.render = () => {
            this.renderer.render(this.scene, this.camera);
        };
        this.renderer = new WebGLRenderer({
            powerPreference: 'high-performance',
            canvas,
            antialias: true,
        });
        this.controls = this.createOrbitControls();
        const near = 0.0001;
        const far = 10000;
        this.setOrbitControls({ minDistance: near, maxDistance: far });
        this.setCameraParameters({
            fov: 45,
            aspect: width / height,
            near,
            far,
            position: {
                x: 0,
                y: 0,
                z: 10,
            },
        });
        this.setSize(width, height);
        this.renderer.setPixelRatio(window.devicePixelRatio);
        this.renderer.setAnimationLoop(this.render);
    }
    // public function
    addPointCloud(cloud, id) {
        this.removePointCloud(id);
        const position = [];
        const color = [];
        const { points } = cloud;
        const size = points.size;
        for (let i = 0; i < size; i++) {
            const point = points.get(i);
            position.push(point.x, point.y, point.z);
            if (point.r !== undefined) {
                color.push(point.r, point.g, point.b);
            }
        }
        const geometry = new BufferGeometry();
        const material = new PointsMaterial();
        if (position.length) {
            geometry.setAttribute('position', new Float32BufferAttribute(position, 3));
        }
        if (color.length) {
            geometry.setAttribute('color', new Float32BufferAttribute(color, 3));
            material.vertexColors = true;
            material.needsUpdate = true;
        }
        this.addPointCloudToScene(new Points(geometry, material), id);
    }
    addPointCloudByUrl(url, id, onProgress) {
        return __awaiter(this, void 0, void 0, function* () {
            this.removePointCloud();
            const cloud = (yield new PCDLoader().loadAsync(url, onProgress));
            if (cloud) {
                this.addPointCloudToScene(cloud, id);
            }
        });
    }
    addPointCloudByData(data, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const cloud = new PCDLoader().parse(data, '');
            this.addPointCloudToScene(cloud, id);
        });
    }
    removePointCloud(id = this.cloudProperties.id) {
        var _a, _b;
        const object = this.scene.getObjectByName(id);
        (_a = object === null || object === void 0 ? void 0 : object.material) === null || _a === void 0 ? void 0 : _a.dispose();
        (_b = object === null || object === void 0 ? void 0 : object.geometry) === null || _b === void 0 ? void 0 : _b.dispose();
        object === null || object === void 0 ? void 0 : object.removeFromParent();
        const index = this.clouds.findIndex((v) => v.name === id);
        this.clouds.splice(index, 1);
    }
    setPointCloudProperties(properties) {
        this.cloudProperties = Object.assign(Object.assign({}, this.cloudProperties), properties);
        const { id } = this.cloudProperties;
        if (!this.clouds.length) {
            return;
        }
        const setProperties = (cloud) => {
            cloud.material.sizeAttenuation = this.cloudProperties.sizeAttenuation;
            cloud.material.size = this.cloudProperties.size;
            cloud.material.color.set(new Color(this.cloudProperties.color));
            cloud.name = id;
        };
        if (id) {
            const object = this.clouds.find((v) => v.name === id);
            if (object) {
                setProperties(object);
            }
        }
        else {
            this.clouds.forEach(setProperties);
        }
    }
    setBackgroundColor(bgc) {
        this.renderer.setClearColor(bgc);
    }
    setCameraParameters(properties) {
        this.camera.fov = properties.fov;
        this.camera.aspect = properties.aspect;
        this.camera.near = properties.near;
        this.camera.far = properties.far;
        if (properties.position) {
            const { x, y, z } = properties.position;
            this.camera.position.set(x, y, z);
        }
    }
    setOrbitControls(properties) {
        var _a;
        for (const key in properties) {
            this.controls[key] = (_a = properties[key]) !== null && _a !== void 0 ? _a : true;
        }
    }
    setAxesHelper(properties) {
        const { visible = true, size = 10 } = properties !== null && properties !== void 0 ? properties : {};
        if (!this.axesHelper) {
            this.axesHelper = new AxesHelper(size);
            this.scene.add(this.axesHelper);
        }
        this.axesHelper.visible = visible;
    }
    setGridHelper(properties) {
        const { visible = true, size = 100, divisions = 100 } = properties !== null && properties !== void 0 ? properties : {};
        if (!this.gridHelper) {
            this.gridHelper = new GridHelper(size, divisions);
            this.scene.add(this.gridHelper);
        }
        this.gridHelper.visible = visible;
    }
    setSize(width, height) {
        this.camera.aspect = width / height;
        this.camera.updateProjectionMatrix();
        this.renderer.setSize(width, height);
    }
    show() {
        this.scene.visible = true;
    }
    hidden() {
        this.scene.visible = false;
    }
    destroy() {
        this.renderer.dispose();
    }
    createOrbitControls() {
        const controls = new OrbitControls(this.camera, this.renderer.domElement);
        return controls;
    }
    addPointCloudToScene(cloud, id) {
        this.scene.add(cloud);
        this.clouds.push(cloud);
        this.setPointCloudProperties({ id });
        this.setOrbitControls({ target: getCenter(cloud.geometry) });
    }
}

export { PointCloudViewer as default };
